/* self-balancing tree routines: splay(), lookup(), insert() */

#include <stdio.h>
#include "sb.h"

struct node *splay(t,k)
/* splays t about key k, returning the new root */
struct node *t;
char *k;
{
    struct node *p = t,*parent,*grandparent;
    char from = ROOT, gpfrom;
    int val;

    if (!t) return NULL;

    /* traverse the tree, stacking nodes */
    while (val = strcmp(k,p->key))
    {
        if (val < 0)
        {
            if (p->left)
            {
                push(p,from);
                from = LEFT;
                p = p->left;
            }
            else break;
        }    
        else
        {
            if (p->right)
            {
                push(p,from);
                from = RIGHT;
                p = p->right;
            }
            else break;
        }
    }

    if (p == t) return p;

    /* now perform rotations on the way back out */
    for(;;)
    {
        /* Case I: p has no grandparent */
        if (fpeek() == ROOT)
        {
            parent = pop();
            if (from == LEFT)
            {
                parent->left = p->right;
                p->right = parent;
            }
            else
            {
                parent->right = p->left;
                p->left = parent;
            }
            return p;
        }
        /* Case II: p is a left child of a left child,
                    or a right child of a right child */
        else if (from == fpeek())
        {
            parent = pop();
            gpfrom = fpeek();
            grandparent = pop();

            if (from == LEFT)
            {
                parent->left = p->right;
                grandparent->left = parent->right;
                p->right = parent;
                parent->right = grandparent;
            }
            else
            {
                parent->right = p->left;
                grandparent->right = parent->left;
                p->left = parent;
                parent->left = grandparent;
            }

            if (grandparent == t) return p;

            /* now wire the whole mess back into the tree */
            if (gpfrom == LEFT) npeek()->left = p;
            else npeek()->right = p;

            from = gpfrom;
        }
        /* Case III: p is a left child of a right child or
                          a right child of a left child */
        else
        {
            parent = pop();
            gpfrom = fpeek();
            grandparent = pop();

            if (from == LEFT)
            {
                parent->left = p->right;
                grandparent->right = p->left;
                p->left = grandparent;
                p->right = parent;
            }
            else
            {
                parent->right = p->left;
                grandparent->left = p->right;
                p->left = parent;
                p->right = grandparent;
            }

            if (grandparent == t) return p;

            /* now wire the whole mess back into the tree */
            if (gpfrom == LEFT) npeek()->left = p;
            else npeek()->right = p;

            from = gpfrom;
        }
    }
}

int lookup(t,k)
/* returns TRUE if k is in t, NULL otherwise */
struct node **t;
char *k;
{
    *t = splay(*t,k);
    return(*t? !strcmp((*t)->key,k) : NULL);
}

int insert(t,k)
/* adds k to t, returns NULL if k already in tree */
struct node **t;
char *k;
{
    struct node *new = (struct node *)malloc(sizeof(struct node));
    char *key = malloc(strlen(k)+1);
    int val;

	new->left = new->right = NULL;
    if (!new)
    {
        fprintf(stderr,"no space for a node\n");
        exit(2);
    }

    strcpy(key,k);
    new->key = key;

    *t = splay(*t,key);

    if (*t)
    {
        if ((val = strcmp(key,(*t)->key))<0)
        {
            new->right = *t;
            new->left = (*t)->left;
            (*t)->left = NULL;
        }
        else if (val > 0)
        {
            new->left = (*t);
            new->right = (*t)->right;
            (*t)->right = NULL;
        }
        else return NULL;
    }    
    *t = new;
    return TRUE;
}
